/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CPU_OPERATOR_CONCATIMPL_FORWARD_KERNEL_H_
#define AIDGE_CPU_OPERATOR_CONCATIMPL_FORWARD_KERNEL_H_

#include <algorithm>
#include <numeric>
#include <cstddef>
#include <vector>

#include "aidge/backend/cpu/operator/ConcatImpl.hpp"
#include "aidge/data/Data.hpp"
#include "aidge/operator/Concat.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"
#include "aidge/backend/cpu/data/GetCPUPtr.h"

namespace Aidge {

template <class I, class O>
void ConcatImpl_cpu_forward_kernel(const Concat_Op::Attrs& attrs,
                                   const std::vector<DimSize_t>& dimsFirstInput,
                                   const std::vector<DimSize_t>& concatAxisValues,
                                   const std::vector<const void*>& inputs_,
                                   void* output_)
{
    // FIXME: missing Concat attributes as arguments
    std::vector<const I*> inputs;
    for (const auto& input_ : inputs_) {
        inputs.push_back(static_cast<const I*>(input_));
    }
    O* output = static_cast<O*>(output_);

    DimSize_t outputAxisValue = std::accumulate(concatAxisValues.begin(), concatAxisValues.end(), 0);

    DimSize_t prodDimLower = 1;
    for (DimIdx_t i = 0; i < std::get<0>(attrs); ++i) {
        prodDimLower *= dimsFirstInput[i];
    }
    DimSize_t prodDimHigher = 1;
    for (DimIdx_t i = std::get<0>(attrs) + 1; static_cast<std::size_t>(i) < dimsFirstInput.size();
         ++i) {
        prodDimHigher *= dimsFirstInput[i];
    }

    std::size_t oIndexStart = 0;
    std::size_t oIndex = 0;
    for (std::size_t inputId = 0; inputId < inputs.size(); ++inputId) {
        oIndex = oIndexStart;
        const DimSize_t iOffset = prodDimHigher*concatAxisValues[inputId];
        for (std::size_t iIndex = 0; iIndex < prodDimLower; ++iIndex) {
            std::copy(inputs[inputId] + iIndex*iOffset, inputs[inputId] + (iIndex+1)*iOffset, output + oIndex);
            oIndex += prodDimHigher*outputAxisValue;
        }
        oIndexStart += concatAxisValues[inputId]*prodDimHigher;
    }
}

namespace {
static Registrar<ConcatImplForward_cpu> registrarConcatImplForward_cpu_Float32(
        {DataType::Float32, DataType::Float32}, Aidge::ConcatImpl_cpu_forward_kernel<float, float>);
static Registrar<ConcatImplForward_cpu> registrarConcatImplForward_cpu_Int32(
        {DataType::Int32, DataType::Int32}, Aidge::ConcatImpl_cpu_forward_kernel<int, int>);
static Registrar<ConcatImplForward_cpu> registrarConcatImplForward_cpu_Float64(
        {DataType::Float64, DataType::Float64},
        Aidge::ConcatImpl_cpu_forward_kernel<double, double>);
}  // namespace
}  // namespace Aidge

#endif /* AIDGE_CPU_OPERATOR_CONCATIMPL_CPU_FORWARD_KERNEL_H_ */
