/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CPU_OPERATOR_SLICEIMPL_FORWARD_KERNEL_H_
#define AIDGE_CPU_OPERATOR_SLICEIMPL_FORWARD_KERNEL_H_

#include <cstddef>
#include <vector>

#include "aidge/backend/cpu/operator/SliceImpl.hpp"
#include "aidge/data/Data.hpp"
#include "aidge/operator/Slice.hpp"
#include "aidge/utils/Registrar.hpp"

namespace Aidge {
template <class I>
void SliceImpl_cpu_forward_kernel(const typename Slice_Op::Attrs& attrs,
                                  const std::vector<std::size_t> inputDims,
                                  const void* input_,
                                  void* output_) {
    std::vector<std::size_t> slicedDims = inputDims;

    std::size_t beginning = 0;
    DimSize_t nbAxes = std::get<2>(attrs).size();
    for (std::size_t i = 0; i < nbAxes; ++i) {
        // For each slice operation get the params and cast them to size_t
        const std::int64_t axis_ = std::get<2>(attrs)[i];
        const std::int64_t start_ = std::get<0>(attrs)[i];
        const std::int64_t end_ = std::get<1>(attrs)[i];
        const std::size_t axis = axis_ >= 0 ? axis_ : static_cast<std::size_t>(axis_) + inputDims.size();
        const std::size_t start = start_ >= 0 ? start_ : start_ + inputDims[axis];
        const std::size_t end = end_ >= 0 ? end_ : end_ + inputDims[axis];
        std::size_t stride = 1;
        for (std::size_t j = inputDims.size() - 1; j > axis; --j) stride *= inputDims[j];
        beginning += start * stride;
        const std::size_t sliceLength = end - start + 1;
        slicedDims[axis] = sliceLength;
    }

    const I* input = static_cast<const I*>(input_) + beginning;
    I* output = static_cast<I*>(output_);
    const std::size_t nbDims = slicedDims.size();

    // for inputDims = {4,5,5,3} & slicedDims = {3,2,2,1}, substractDims = {1,5,5,3}
    std::vector<std::size_t> substractedDims = std::vector<std::size_t>(nbDims);
    for (std::size_t i = 0; i < nbDims; ++i) {
        substractedDims[i] = inputDims[i] - slicedDims[i];
    }

    // for slicedDims = {3,2,2,1}, prodSlicedDims = {12,4,2,1}
    std::vector<std::size_t> prodSlicedDims = std::vector<std::size_t>(nbDims);
    std::vector<std::size_t> prodInputDims = std::vector<std::size_t>(nbDims + 1);
    prodSlicedDims[nbDims - 1] = slicedDims[nbDims - 1];
    prodInputDims[nbDims - 1] = inputDims[nbDims - 1];
    prodInputDims[nbDims] = 1;
    for (std::size_t i = 2; i <= nbDims; ++i) {
        prodSlicedDims[nbDims - i] = prodSlicedDims[nbDims - i + 1] * slicedDims[nbDims - i];
        prodInputDims[nbDims - i] = prodInputDims[nbDims - i + 1] * inputDims[nbDims - i];
    }

    std::size_t j = 0;
    std::size_t i = 0;
    for (; j < prodSlicedDims[0];) {
        output[j] = input[i++];
        ++j;
        for (std::size_t idx = nbDims - 1; idx > 0; --idx) {
            i += j % prodSlicedDims[idx] == 0 ? substractedDims[idx] * prodInputDims[idx + 1] : 0;
        }
    }
}

namespace {

static Registrar<SliceImplForward_cpu> registrarSliceImplForward_cpu_Float32(
        {DataType::Float32}, Aidge::SliceImpl_cpu_forward_kernel<float>);
static Registrar<SliceImplForward_cpu> registrarSliceImplForward_cpu_Int32(
        {DataType::Int32}, Aidge::SliceImpl_cpu_forward_kernel<int>);
static Registrar<SliceImplForward_cpu> registrarSliceImplForward_cpu_Float64(
        {DataType::Float64}, Aidge::SliceImpl_cpu_forward_kernel<double>);
}  // namespace
}  // namespace Aidge

#endif /* AIDGE_CPU_OPERATOR_SLICEIMPL_FORWARD_KERNEL_H_ */
